"""Asynchronous Python client for the AdGuard Home API."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .exceptions import AdGuardHomeError

if TYPE_CHECKING:
    from . import AdGuardHome


@dataclass
class AdGuardHomeAvailableUpdate:
    """Latest available AdGuard Home update."""

    disabled: bool
    new_version: str | None = None
    announcement: str | None = None
    announcement_url: str | None = None
    can_autoupdate: bool | None = None


@dataclass
class AdGuardHomeUpdate:
    """Controls AdGuard Home version update."""

    adguard: AdGuardHome

    async def update_available(self) -> AdGuardHomeAvailableUpdate:
        """Return AdGuard Home latest available update.

        Returns
        -------
            An AdGuardHomeAvailableUpdate object with all data about the latest update.

        """
        response = await self.adguard.request("version.json", method="POST")
        return AdGuardHomeAvailableUpdate(**response)

    async def begin_update(self) -> None:
        """Begin AdGuard Home auto-upgrade procedure.

        Raises
        ------
            AdGuardHomeError: If beginning the auto-upgrade failed.

        """
        try:
            await self.adguard.request("update", method="POST")
        except AdGuardHomeError as exception:
            msg = "Beginning AdGuard Home update failed"
            raise AdGuardHomeError(msg) from exception



"""Asynchronous Python client for the AdGuard Home API."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .exceptions import AdGuardHomeError

if TYPE_CHECKING:
    from . import AdGuardHome


@dataclass
class AdGuardHomeSafeSearch:
    """Controls AdGuard Home safe search enforcing."""

    adguard: AdGuardHome

    async def enabled(self) -> bool:
        """Return if AdGuard Home safe search enforcing is enabled or not.

        Returns
        -------
            The current state of the AdGuard Home safe search.

        """
        response = await self.adguard.request("safesearch/status")
        return response["enabled"]

    async def enable(self) -> None:
        """Enable AdGuard Home safe search enforcing.

        Raises
        ------
            AdGuardHomeError: If enabling the safe search didn't succeed.

        """
        try:
            await self.adguard.request("safesearch/enable", method="POST")
        except AdGuardHomeError as exception:
            msg = "Enabling AdGuard Home safe search failed"
            raise AdGuardHomeError(msg) from exception

    async def disable(self) -> None:
        """Disable AdGuard Home safe search enforcing.

        Raises
        ------
            AdGuardHomeError: If disabling the safe search didn't succeed.

        """
        try:
            await self.adguard.request("safesearch/disable", method="POST")
        except AdGuardHomeError as exception:
            msg = "Disabling AdGuard Home safe search failed"
            raise AdGuardHomeError(msg) from exception



"""Asynchronous Python client for the AdGuard Home API."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .exceptions import AdGuardHomeError

if TYPE_CHECKING:
    from . import AdGuardHome


@dataclass
class AdGuardHomeParental:
    """Controls AdGuard Home parental control."""

    adguard: AdGuardHome

    async def enabled(self) -> bool:
        """Return if AdGuard Home parental control is enabled or not.

        Returns
        -------
            The current state of the AdGuard Home parental control.

        """
        response = await self.adguard.request("parental/status")
        return response["enabled"]

    async def enable(self) -> None:
        """Enable AdGuard Home parental control.

        Raises
        ------
            AdGuardHomeError: If enabling parental control failed.

        """
        try:
            await self.adguard.request("parental/enable", method="POST")
        except AdGuardHomeError as exception:
            msg = "Enabling AdGuard Home parental control failed"
            raise AdGuardHomeError(msg) from exception

    async def disable(self) -> None:
        """Disable AdGuard Home parental control.

        Raises
        ------
            AdGuardHomeError: If disabling parental control failed.

        """
        try:
            await self.adguard.request("parental/disable", method="POST")
        except AdGuardHomeError as exception:
            msg = "Disabling AdGuard Home parental control failed"
            raise AdGuardHomeError(msg) from exception



"""Asynchronous Python client for the AdGuard Home API."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .exceptions import AdGuardHomeError

if TYPE_CHECKING:
    from . import AdGuardHome


@dataclass
class AdGuardHomeQueryLog:
    """Controls AdGuard Home query log."""

    adguard: AdGuardHome

    async def _config(
        self, *, enabled: bool | None = None, interval: int | None = None
    ) -> None:
        """Configure query log on AdGuard Home.

        Args:
        ----
            enabled: Enable/disable AdGuard Home query log.
            interval: Number of day to keep data in the logs.

        """
        if enabled is None:
            enabled = await self.enabled()
        if interval is None:
            interval = await self.interval()
        await self.adguard.request(
            "querylog_config",
            method="POST",
            json_data={"enabled": enabled, "interval": interval},
        )

    async def enabled(self) -> bool:
        """Return if AdGuard Home query log is enabled or not.

        Returns
        -------
            The current state of the AdGuard Home query log.

        """
        response = await self.adguard.request("querylog_info")
        return response["enabled"]

    async def enable(self) -> None:
        """Enable AdGuard Home query log.

        Raises
        ------
            AdGuardHomeError: If enabling the query log didn't succeed.

        """
        try:
            await self._config(enabled=True)
        except AdGuardHomeError as exception:
            msg = "Enabling AdGuard Home query log failed"
            raise AdGuardHomeError(msg) from exception

    async def interval(self, interval: int | None = None) -> int:
        """Return or set the time period to keep query log data.

        Args:
        ----
            interval: Set the time period (in days) to keep query log data.

        Returns:
        -------
            The current set time period to keep query log data.

        """
        if interval:
            await self._config(interval=interval)
            return interval

        response = await self.adguard.request("querylog_info")
        return response["interval"]

    async def disable(self) -> None:
        """Disable AdGuard Home query log.

        Raises
        ------
            AdGuardHomeError: If disabling the query filter log didn't succeed.

        """
        try:
            await self._config(enabled=False)
        except AdGuardHomeError as exception:
            msg = "Disabling AdGuard Home query log failed"
            raise AdGuardHomeError(msg) from exception



"""Asynchronous Python client for the AdGuard Home API."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .exceptions import AdGuardHomeError

if TYPE_CHECKING:
    from . import AdGuardHome


@dataclass
class AdGuardHomeStats:
    """Provides stats of AdGuard Home."""

    adguard: AdGuardHome

    async def dns_queries(self) -> int:
        """Return number of DNS queries.

        Returns
        -------
            The number of DNS queries performed by the AdGuard Home instance.

        """
        response = await self.adguard.request("stats")
        return response["num_dns_queries"]

    async def blocked_filtering(self) -> int:
        """Return number of blocked DNS queries.

        Returns
        -------
            The number of DNS queries blocked by the AdGuard Home instance.

        """
        response = await self.adguard.request("stats")
        return response["num_blocked_filtering"]

    async def blocked_percentage(self) -> float:
        """Return the blocked percentage ratio of DNS queries.

        Returns
        -------
            The percentage ratio of blocked DNS queries by the AdGuard Home
            instance.

        """
        response = await self.adguard.request("stats")
        if not response["num_dns_queries"]:
            return 0.0
        return (response["num_blocked_filtering"] / response["num_dns_queries"]) * 100.0

    async def replaced_safebrowsing(self) -> int:
        """Return number of blocked pages by safe browsing.

        Returns
        -------
            The number of times a page was blocked by the safe
            browsing feature of the AdGuard Home instance.

        """
        response = await self.adguard.request("stats")
        return response["num_replaced_safebrowsing"]

    async def replaced_parental(self) -> int:
        """Return number of blocked pages by parental control.

        Returns
        -------
            The number of times a page was blocked by the parental control
            feature of the AdGuard Home instance.

        """
        response = await self.adguard.request("stats")
        return response["num_replaced_parental"]

    async def replaced_safesearch(self) -> int:
        """Return number of enforced safe searches.

        Returns
        -------
            The number of times a safe search was enforced by the
            AdGuard Home instance.

        """
        response = await self.adguard.request("stats")
        return response["num_replaced_safesearch"]

    async def avg_processing_time(self) -> float:
        """Return average processing time of DNS queries (in ms).

        Returns
        -------
            The averages processing time (in milliseconds) of DNS queries
            as performed by the AdGuard Home instance.

        """
        response = await self.adguard.request("stats")
        return round(response["avg_processing_time"] * 1000, 2)

    async def period(self) -> int:
        """Return the time period to keep data (in days).

        Returns
        -------
            The time period of data this AdGuard Home instance keeps.

        """
        response = await self.adguard.request("stats_info")
        return response["interval"]

    async def reset(self) -> None:
        """Reset all stats.

        Raises
        ------
            AdGuardHomeError: Restting the AdGuard Home stats did not succeed.

        """
        try:
            await self.adguard.request("stats_reset", method="POST")
        except AdGuardHomeError as exception:
            msg = "Resetting AdGuard Home stats failed"
            raise AdGuardHomeError(msg) from exception



"""Asynchronous Python client for the AdGuard Home API."""

from .adguardhome import AdGuardHome
from .exceptions import AdGuardHomeConnectionError, AdGuardHomeError

__all__ = ["AdGuardHome", "AdGuardHomeConnectionError", "AdGuardHomeError"]



"""Asynchronous Python client for the AdGuard Home API."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .exceptions import AdGuardHomeError

if TYPE_CHECKING:
    from . import AdGuardHome


@dataclass
class AdGuardHomeFiltering:
    """Controls AdGuard Home filtering. Blocks domains."""

    adguard: AdGuardHome

    async def _config(
        self, *, enabled: bool | None = None, interval: int | None = None
    ) -> None:
        """Configure filtering on AdGuard Home.

        Args:
        ----
            enabled: Enable/Disable AdGuard Home filtering.
            interval: Number of days to keep data in the logs.

        """
        if enabled is None:
            enabled = await self.enabled()
        if interval is None:
            interval = await self.interval()

        await self.adguard.request(
            "filtering/config",
            method="POST",
            json_data={"enabled": enabled, "interval": interval},
        )

    async def enabled(self) -> bool:
        """Return if AdGuard Home filtering is enabled or not.

        Returns
        -------
            The current state of the AdGuard Home filtering.

        """
        response = await self.adguard.request("filtering/status")
        return response["enabled"]

    async def enable(self) -> None:
        """Enable AdGuard Home filtering.

        Raises
        ------
            AdGuardHomeError: If enabling the filtering didn't succeed.

        """
        try:
            await self._config(enabled=True)
        except AdGuardHomeError as exception:
            msg = "Enabling AdGuard Home filtering failed"
            raise AdGuardHomeError(msg) from exception

    async def disable(self) -> None:
        """Disable AdGuard Home filtering.

        Raises
        ------
            AdGuardHomeError: If disabling the filtering didn't succeed.

        """
        try:
            await self._config(enabled=False)
        except AdGuardHomeError as exception:
            msg = "Disabling AdGuard Home filtering failed"
            raise AdGuardHomeError(msg) from exception

    async def interval(self, *, interval: int | None = None) -> int:
        """Return or set the time period to keep query log data.

        Args:
        ----
            interval: Set the time period (in days) to keep query log data.

        Returns:
        -------
            The current set time period to keep query log data.

        """
        if interval:
            await self._config(interval=interval)
            return interval

        response = await self.adguard.request("filtering/status")
        return response["interval"]

    async def rules_count(self, *, allowlist: bool) -> int:
        """Return the number of rules loaded.

        Args:
        ----
            allowlist: True to get the allowlists count, False for the blocklists.

        Returns:
        -------
            The number of filtering rules currently loaded in the AdGuard
            Home instance.

        """
        response = await self.adguard.request("filtering/status")

        count = "whitelist_filters" if allowlist else "filters"
        if not response.get(count):
            return 0

        return sum(fil["rules_count"] for fil in response[count])

    async def add_url(self, *, allowlist: bool, name: str, url: str) -> None:
        """Add a new filter subscription to AdGuard Home.

        Args:
        ----
            allowlist: True to add an allowlist, False for a blocklists.
            name: The name of the filter subscription.
            url: The URL of the filter list.

        Raises:
        ------
            AdGuardHomeError: Failed adding the filter subscription.

        """
        try:
            await self.adguard.request(
                "filtering/add_url",
                method="POST",
                json_data={"whitelist": allowlist, "name": name, "url": url},
            )
        except AdGuardHomeError as exception:
            msg = "Failed adding URL to AdGuard Home filter"
            raise AdGuardHomeError(msg) from exception

    async def remove_url(self, *, allowlist: bool, url: str) -> None:
        """Remove a new filter subscription from AdGuard Home.

        Args:
        ----
            allowlist: True to remove an allowlist, False for a blocklists.
            url: Filter subscription URL to remove from AdGuard Home.

        Raises:
        ------
            AdGuardHomeError: Failed removing the filter subscription.

        """
        try:
            await self.adguard.request(
                "filtering/remove_url",
                method="POST",
                json_data={"whitelist": allowlist, "url": url},
            )
        except AdGuardHomeError as exception:
            msg = "Failed removing URL from AdGuard Home filter"
            raise AdGuardHomeError(msg) from exception

    async def enable_url(self, *, allowlist: bool, url: str) -> None:
        """Enable a filter subscription in AdGuard Home.

        Args:
        ----
            allowlist: True to enable an allowlist, False for a blocklists.
            url: Filter subscription URL to enable on AdGuard Home.

        Raises:
        ------
            AdGuardHomeError: Failed enabling filter subscription.

        """
        response = await self.adguard.request("filtering/status")
        filter_type = "whitelist_filters" if allowlist else "filters"

        # Excluded from coverage:
        # https://github.com/nedbat/coveragepy/issues/515
        name = next(  # pragma: no cover
            (
                fil["name"]
                for fil in response[filter_type]
                if fil["url"].lower() == url.lower()
            ),
            "Unknown",
        )

        try:
            await self.adguard.request(
                "filtering/set_url",
                method="POST",
                json_data={
                    "url": url,
                    "whitelist": allowlist,
                    "data": {"enabled": True, "name": name, "url": url},
                },
            )
        except AdGuardHomeError as exception:
            msg = "Failed enabling URL on AdGuard Home filter"
            raise AdGuardHomeError(msg) from exception

    async def disable_url(self, *, allowlist: bool, url: str) -> None:
        """Disable a filter subscription in AdGuard Home.

        Args:
        ----
            url: Filter subscription URL to disable on AdGuard Home.
            allowlist: True to update the allowlists, False for the blocklists.

        Raises:
        ------
            AdGuardHomeError: Failed disabling filter subscription.

        """
        response = await self.adguard.request("filtering/status")
        filter_type = "whitelist_filters" if allowlist else "filters"

        # Excluded from coverage:
        # https://github.com/nedbat/coveragepy/issues/515
        name = next(  # pragma: no cover
            (
                fil["name"]
                for fil in response[filter_type]
                if fil["url"].lower() == url.lower()
            ),
            "Unknown",
        )

        try:
            await self.adguard.request(
                "filtering/set_url",
                method="POST",
                json_data={
                    "url": url,
                    "whitelist": allowlist,
                    "data": {"enabled": False, "name": name, "url": url},
                },
            )
        except AdGuardHomeError as exception:
            msg = "Failed disabling URL on AdGuard Home filter"
            raise AdGuardHomeError(msg) from exception

    async def url_enabled(self, *, allowlist: bool, url: str) -> bool:
        """Check if a filter subscription is enabled in AdGuard Home.

        Args:
        ----
            allowlist: True to check an allowlist, False for a blocklists.
            url: Filter subscription URL to check on AdGuard Home.

        Returns:
        -------
            True if the filter subscription is enabled, False otherwise.

        """
        response = await self.adguard.request("filtering/status")
        filter_type = "whitelist_filters" if allowlist else "filters"
        filters = response.get(filter_type) or []

        return next(
            (fil["enabled"] for fil in filters if fil["url"].lower() == url.lower()),
            False,
        )

    async def refresh(self, *, allowlist: bool, force: bool = False) -> None:
        """Reload filtering subscriptions from URLs specified in AdGuard Home.

        Args:
        ----
            force: Force the reload of all filter subscriptions.
            allowlist: True to update the allowlists, False for the blocklists.

        Raises:
        ------
            AdGuardHomeError: Failed to refresh filter subscriptions.

        """
        force_value = "true" if force else "false"

        try:
            await self.adguard.request(
                "filtering/refresh",
                method="POST",
                json_data={"whitelist": allowlist},
                params={"force": force_value},
            )
        except AdGuardHomeError as exception:
            msg = "Failed refreshing filter URLs in AdGuard Home"
            raise AdGuardHomeError(msg) from exception



"""Asynchronous Python client for the AdGuard Home API."""

from __future__ import annotations

import asyncio
import json
import socket
from typing import TYPE_CHECKING, Any, Self

import aiohttp
from yarl import URL

from .exceptions import AdGuardHomeConnectionError, AdGuardHomeError
from .filtering import AdGuardHomeFiltering
from .parental import AdGuardHomeParental
from .querylog import AdGuardHomeQueryLog
from .safebrowsing import AdGuardHomeSafeBrowsing
from .safesearch import AdGuardHomeSafeSearch
from .stats import AdGuardHomeStats
from .update import AdGuardHomeUpdate

if TYPE_CHECKING:
    from collections.abc import Mapping


# pylint: disable=too-many-instance-attributes
class AdGuardHome:
    """Main class for handling connections with AdGuard Home."""

    # pylint: disable-next=too-many-arguments
    def __init__(  # noqa: PLR0913
        self,
        host: str,
        *,
        base_path: str = "/control",
        password: str | None = None,
        port: int = 3000,
        request_timeout: int = 10,
        session: aiohttp.client.ClientSession | None = None,
        tls: bool = False,
        username: str | None = None,
        verify_ssl: bool = True,
    ) -> None:
        """Initialize connection with AdGuard Home.

        Class constructor for setting up an AdGuard Home object to
        communicate with an AdGuard Home instance.

        Args:
        ----
            host: Hostname or IP address of the AdGuard Home instance.
            base_path: Base path of the API, usually `/control`, which is the default.
            password: Password for HTTP auth, if enabled.
            port: Port on which the API runs, usually 3000.
            request_timeout: Max timeout to wait for a response from the API.
            session: Optional, shared, aiohttp client session.
            tls: True, when TLS/SSL should be used.
            username: Username for HTTP auth, if enabled.
            verify_ssl: Can be set to false, when TLS with self-signed cert is used.

        """
        self._session = session
        self._close_session = False

        self.base_path = base_path
        self.host = host
        self.password = password
        self.port = port
        self.request_timeout = request_timeout
        self.tls = tls
        self.username = username
        self.verify_ssl = verify_ssl

        if self.base_path[-1] != "/":
            self.base_path += "/"

        self.filtering = AdGuardHomeFiltering(self)
        self.parental = AdGuardHomeParental(self)
        self.querylog = AdGuardHomeQueryLog(self)
        self.safebrowsing = AdGuardHomeSafeBrowsing(self)
        self.safesearch = AdGuardHomeSafeSearch(self)
        self.stats = AdGuardHomeStats(self)
        self.update = AdGuardHomeUpdate(self)

    # pylint: disable-next=too-many-arguments, too-many-locals, too-many-positional-arguments
    async def request(
        self,
        uri: str,
        method: str = "GET",
        data: Any | None = None,
        json_data: dict[str, Any] | None = None,
        params: Mapping[str, str] | None = None,
    ) -> dict[str, Any]:
        """Handle a request to the AdGuard Home instance.

        Make a request against the AdGuard Home API and handles the response.

        Args:
        ----
            uri: The request URI on the AdGuard Home API to call.
            method: HTTP method to use for the request; e.g., GET, POST.
            data: RAW HTTP request data to send with the request.
            json_data: Dictionary of data to send as JSON with the request.
            params: Mapping of request parameters to send with the request.

        Returns:
        -------
            The response from the API. In case the response is a JSON response,
            the method will return a decoded JSON response as a Python
            dictionary. In other cases, it will return the RAW text response.

        Raises:
        ------
            AdGuardHomeConnectionError: An error occurred while communicating
                with the AdGuard Home instance (connection issues).
            AdGuardHomeError: An error occurred while processing the
                response from the AdGuard Home instance (invalid data).

        """
        scheme = "https" if self.tls else "http"
        url = URL.build(
            scheme=scheme, host=self.host, port=self.port, path=self.base_path
        ).join(URL(uri))

        auth = None
        if self.username and self.password:
            auth = aiohttp.BasicAuth(self.username, self.password)

        headers = {
            "Accept": "application/json, text/plain, */*",
        }

        if self._session is None:
            self._session = aiohttp.ClientSession()
            self._close_session = True

        skip_auto_headers = None
        if data is None and json_data is None:
            skip_auto_headers = {"Content-Type"}

        try:
            async with asyncio.timeout(self.request_timeout):
                response = await self._session.request(
                    method,
                    url,
                    auth=auth,
                    data=data,
                    json=json_data,
                    params=params,
                    headers=headers,
                    ssl=self.verify_ssl,
                    skip_auto_headers=skip_auto_headers,
                )
        except TimeoutError as exception:
            msg = "Timeout occurred while connecting to AdGuard Home instance."
            raise AdGuardHomeConnectionError(msg) from exception
        except (aiohttp.ClientError, socket.gaierror) as exception:
            msg = "Error occurred while communicating with AdGuard Home."
            raise AdGuardHomeConnectionError(msg) from exception

        content_type = response.headers.get("Content-Type", "")
        if response.status // 100 in [4, 5]:
            contents = await response.read()
            response.close()

            if content_type == "application/json":
                raise AdGuardHomeError(
                    response.status, json.loads(contents.decode("utf8"))
                )
            raise AdGuardHomeError(
                response.status, {"message": contents.decode("utf8")}
            )

        if "application/json" in content_type:
            return await response.json()

        text = await response.text()
        return {"message": text}

    async def protection_enabled(self) -> bool:
        """Return if AdGuard Home protection is enabled or not.

        Returns
        -------
            The status of the protection of the AdGuard Home instance.

        """
        response = await self.request("status")
        return response["protection_enabled"]

    async def enable_protection(self) -> None:
        """Enable AdGuard Home protection.

        Raises
        ------
            AdGuardHomeError: Failed enabling AdGuard Home protection.

        """
        try:
            await self.request(
                "dns_config",
                method="POST",
                json_data={"protection_enabled": True},
            )
        except AdGuardHomeError as exception:
            msg = "Failed enabling AdGuard Home protection"
            raise AdGuardHomeError(msg) from exception

    async def disable_protection(self) -> None:
        """Disable AdGuard Home protection.

        Raises
        ------
            AdGuardHomeError: Failed disabling the AdGuard Home protection.

        """
        try:
            await self.request(
                "dns_config",
                method="POST",
                json_data={"protection_enabled": False},
            )
        except AdGuardHomeError as exception:
            msg = "Failed disabling AdGuard Home protection"
            raise AdGuardHomeError(msg) from exception

    async def version(self) -> str:
        """Return the current version of the AdGuard Home instance.

        Returns
        -------
            The version number of the connected AdGuard Home instance.

        """
        response = await self.request("status")
        return response["version"]

    async def close(self) -> None:
        """Close open client session."""
        if self._session and self._close_session:
            await self._session.close()

    async def __aenter__(self) -> Self:
        """Async enter.

        Returns
        -------
            The AdGuard Home object.

        """
        return self

    async def __aexit__(self, *_exc_info: object) -> None:
        """Async exit.

        Args:
        ----
            _exc_info: Exec type.

        """
        await self.close()
